/*
 * TunnelingRMIAgent.java
 *
 * Copyright 2007 Sun Microsystems, Inc.  All Rights Reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 *   - Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *
 *   - Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution.
 *
 *   - Neither the name of Sun Microsystems nor the names of its
 *     contributors may be used to endorse or promote products derived
 *     from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
 * IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
 * THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE COPYRIGHT OWNER OR
 * CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * Created on Jul 25, 2007, 11:42:49 AM
 *
 */

package edu.ucmerced.jmx;

import java.io.IOException;
import java.lang.management.ManagementFactory;
import java.lang.instrument.Instrumentation;
import java.net.InetAddress;
import java.rmi.registry.LocateRegistry;
import java.util.HashMap;
import javax.management.MBeanServer;
import javax.management.remote.JMXConnectorServer;
import javax.management.remote.JMXConnectorServerFactory;
import javax.management.remote.JMXServiceURL;


/**
 * This TunnelingRMIAgent will start an RMI Connector Server using only
 * port "edu.ucmerced.jmx.customagent.port".
 *
 * @author dfuchs
 */
public class TunnelingRMIAgent {
    /* Not using Logger because it always seems to print twice :-( */

    private static final String DEFAULT_RMI_PORT = "1100";

    private static JMXConnectorServer cs;

    private TunnelingRMIAgent() { }

    public static void premain(String agentArgs, Instrumentation inst)
            throws IOException {
        premain(agentArgs);
    }

    public static void premain(String agentArgs)
            throws IOException {

        final int port = Integer.parseInt(
                System.getProperty("edu.ucmerced.jmx.tunnelingrmiagent.port",
                DEFAULT_RMI_PORT));
        System.out.println("Starting tunneling RMI agent on port " + port);

        // Ensure cryptographically strong random number generator used
        // to choose the object number - see java.rmi.server.ObjID
        //
        System.setProperty("java.rmi.server.randomIDs", "true");

        LocateRegistry.createRegistry(port);

        // Retrieve the PlatformMBeanServer.
        //
        MBeanServer mbs = ManagementFactory.getPlatformMBeanServer();

        // Environment map.
        //
        HashMap<String, Object> env = new HashMap<String, Object>();

        // This where we would enable security - left out of this
        // for the sake of the example....
        //

        // Create an RMI connector server.
        //
        // As specified in the JMXServiceURL the RMIServer stub will be
        // registered in the RMI registry running in the local host on
        // port 1100 with the name "jmxrmi". This is the same name the
        // out-of-the-box management agent uses to register the RMIServer
        // stub too.
        //
        // The port specified in "service:jmx:rmi://"+hostname+":"+port
        // is the second port, where RMI connection objects will be exported.
        // Here we use the same port as that we choose for the RMI registry.
        // The port for the RMI registry is specified in the second part
        // of the URL, in "rmi://"+hostname+":"+port
        //
        //final String hostname = InetAddress.getLocalHost().getHostName();
        final String hostname = "localhost";
        String urls = "service:jmx:rmi://" + hostname +
            ":" + port + "/jndi/rmi://" + hostname + ":" + port + "/jmxrmi";
        JMXServiceURL url = new JMXServiceURL(urls);

        // Now create the server from the JMXServiceURL
        //
        cs = JMXConnectorServerFactory.newJMXConnectorServer(url, env, mbs);

        // Start the RMI connector server.
        //
        cs.start();

        // Terminate after the main thread finishes.
        //
        final Thread main = getMainThread();
        if (main == null) {
            System.out.println("WARN: Cannot identify main thread.");
            System.out.println("WARN: The tunneling RMI agent will not terminate properly.");
        } else {
            final Thread clean = new Thread("TunnelingRMIAgent cleaner") {
                public void run() {
                    try {
                        main.join();
                    } catch (Exception ex) {
                        ex.printStackTrace();
                    } finally {
                        try {
                            System.out.println("Stopping the tunneling RMI agent.");
                            cs.stop();
                        } catch (IOException ex) {
                            // FIXME: It always throws an I/O exception.
                        } catch (Exception ex) {
                            ex.printStackTrace();
                        }
                    }
                }
            };
            clean.setDaemon(true);
            clean.start();
        }
    }

    private static Thread getMainThread() {
        final Thread current = Thread.currentThread();
        if (current.getName().equals("main")) return current;

        final Thread[] all = new Thread[Thread.activeCount() + 100];
        final int count = Thread.enumerate(all);

        for (int i = 0; i < count; i++) {
            if (all[i].getName().equals("main")) {
                return all[i];
            }
        }
        return null;
    }

}

